#!/usr/bin/env/python
# -*- coding: utf-8 -*-

''' Module for the command line interface to data-fetcher. '''

import logging
import sys
from argparse import Namespace
from pathlib import Path

from datafetcher.project_fetcher import ParallelFetcher
from datafetcher.return_codes import ReturnCode

from .locations_report import LocationsReport
from .utilities import get_arg_parser, get_capo_settings, FlexLogger, \
    path_is_accessible

# pylint: disable=W1203, R0902, R0903

_APPLICATION_NAME = 'datafetcher'
_VALID_PROFILES = ['naasc-dev', 'naasc-test', 'dsoc-test', 'dsoc-dev',
                   'nmprod', 'local']

class DataFetcher:
    '''
        TO EXECUTE ALL DF TESTS: from datafetcher/, just run pytest

        example command line that should work with the correct local profile
        and a current product locator (these are changed whenever the
        database is rebuilt:
        datafetcher --profile local --output-dir ~/Downloads/
        --product-locator \
            uid://evla/execblock/93e1c0cd-76de-4e65-a3f2-d5fe55f386d8 \
            --verbose

        local.properties must have:
        - edu.nrao.archive.workflow.config
            .StartupSettings.temporaryDataDirectory pointing to a locally
            writable temp dir, e.g., /var/tmp
        - edu.nrao.archive.workflow.config.DeliverySettings.hostname must point
            to local computer
        - execution_site must NOT be DSOC or NAASC

    '''

    def __init__(self, args: Namespace, df_capo_settings: dict):
        self.usage = self._build_usage_message()
        if args is None or df_capo_settings is None:
            self._exit_with_error(ReturnCode.MISSING_SETTING)
        self.args = args
        self.settings = df_capo_settings
        self.verbose = self.args.verbose

        # required arguments
        self.output_dir = args.output_dir
        if self.output_dir is None:
            self._exit_with_error(ReturnCode.MISSING_SETTING)

        output_dir = Path(self.output_dir)
        if not output_dir.is_dir()\
                or not path_is_accessible(output_dir):
            logging.error(f'output location {self.output_dir} inaccessible '
                          f'or not found')
            self._exit_with_error(ReturnCode.MISSING_SETTING)

        self._LOG = FlexLogger(self.__class__.__name__,
                               self.output_dir, self.verbose)
        if args.location_file is not None:
            if args.product_locator is not None:
                self._LOG.error('required: location file OR product locator '
                                '-- not both')
                self._exit_with_error(ReturnCode.MISSING_SETTING)
            self.location_file = args.location_file
        elif args.product_locator is not None:
            self.product_locator = args.product_locator
        else:
            self._LOG.error('you must specify either a location file or a '
                            'product locator')
            self._exit_with_error(ReturnCode.MISSING_SETTING)

        self.profile = args.profile
        if self.profile is None or not self.profile in _VALID_PROFILES:
            self._exit_with_error(ReturnCode.MISSING_PROFILE)

        # optional arguments
        self.is_dry = args.dry_run
        self.force  = args.force
        self.verbose = args.verbose or False

        try:
            self.locations_report = self._get_locations()
            self.servers_report = self.locations_report.servers_report
        except SystemExit as exc:
            self._LOG.error(f'{exc}')
            if args.location_file:
                self._exit_with_error(ReturnCode.CANNOT_OPEN_LOCATION_FILE)
            else:
                self._exit_with_error(ReturnCode.PRODUCT_LOCATOR_NOT_FOUND)
            raise

        except Exception as exc:
            self._LOG.error(
                f'>>> throwing unexpected {type(exc)} during init: {exc}')
            raise

    def _exit_with_error(self, return_code: ReturnCode):
        print(self.usage)
        sys.exit(return_code.value['code'])


    @staticmethod
    def _build_usage_message() -> str:
        usage_str = 'Usage:'
        usage_str += """\tdatafetcher [-h]
             (--product-locator PRODUCT_LOCATOR | --location-file LOCATION_FILE)
             [--dry-run] [--output-dir OUTPUT_DIR] [-v, --verbose]
             [--profile PROFILE]\n"""
        usage_str += "\n\t\tReturn codes:\n"
        for return_code in ReturnCode:
            usage_str += f"\n\t\t\t{return_code.value['code']}: " \
                         f"{return_code.value['text']}"
        return usage_str

    def run(self):
        """
        launch the fetcher
        :return:
        """

        try:
            fetcher = ParallelFetcher(
                self.args, self.settings, self._LOG,
                self.servers_report)
            return fetcher.run()
        except SystemExit as exc:
            self._LOG.error(f'{exc}')
            raise
        except Exception as exc:
            self._LOG.error(
                f'>>> throwing unexpected exception during run: {exc}')
            raise

    def _get_locations(self):
        try:
            capo_settings = get_capo_settings(self.profile)
            return LocationsReport(self._LOG,
                                   self.args,
                                   capo_settings)
        except Exception as exc:
            self._LOG.error(f'{exc}')
            if hasattr(self, 'location_file'):
                self._exit_with_error(ReturnCode.CANNOT_OPEN_LOCATION_FILE)
            elif hasattr(self, 'product_locator'):
                self._exit_with_error(ReturnCode.PRODUCT_LOCATOR_NOT_FOUND)
            else:
                # should never happen; we've already checked
                self._exit_with_error(ReturnCode.MISSING_SETTING)


def main():
    ''' this will be executed when fetcher is launched
        from the command line
    '''

    parser = get_arg_parser()
    try:
        args = parser.parse_args()
    except Exception as exc:
        logging.error(f'{exc}')
        return exc.value
    except SystemExit as exc:
        return exc.value.code
    settings = get_capo_settings(args.profile)
    datafetcher = DataFetcher(args, settings)
    return datafetcher.run()


if __name__ == '__main__':
    main()
