import logging
import pendulum
import subprocess
from pathlib import Path
from typing import NamedTuple, List, Dict
from collections import namedtuple

from workspaces.schema import WorkflowEventType


class MockLogger:
    """
    Class representing a logger for Vulture that will simulate HTCondor logs
    """

    def __init__(self, log_path: Path):
        self.logger = logging.getLogger("vulture")
        self.logger.setLevel(logging.INFO)
        self.logger.addHandler(logging.FileHandler(f"{log_path}", mode="a+"))

    def log_event(self, event_type: WorkflowEventType, event_body: str = None):
        """
        Log a mock HTCondor event to the log file
        :param event_type: Type of the event
        :param event_body: Body of the event log (can be empty)
        :return:
        """
        header = self.create_header(event_type)
        # Indent each line of the body
        if event_body:
            event_body_list = event_body.split("\n")
            if len(event_body_list) > 1:
                event_body = "\t".join(event_body_list)
            else:
                event_body = "\t" + event_body
        footer = "..."
        self.logger.info(
            msg="\n".join([header, event_body, footer])
        ) if event_body else self.logger.info("\n".join([header, footer]))

    @staticmethod
    def create_header(event_type: WorkflowEventType) -> str:
        """
        Create mock HTCondor event log header
        :param event_type: Type of event being logged
        :return: The header as a string
        EXAMPLE:
        000 (3983.000.000) 08/26 11:06:06 Job submitted from host: <10.64.1.178:9618?addrs=10.64.1.178-9618&noUDP&sock=4050180_7e37_3>
        """
        mock_host = "<0.0.0.0:7777>"
        event_descs = {
            WorkflowEventType.SUBMITTED: f"Job submitted from host: {mock_host}",
            WorkflowEventType.EXECUTING: f"Job executing on host: {mock_host}",
            WorkflowEventType.TERMINATED: "Job terminated.",
            WorkflowEventType.OTHER: "Unknown description.",
        }
        event_num = f"{event_type.value:03d}"
        # TODO: Add calculable job number
        job_num = f"({4242:04d}.000.000)"
        timestamp = f"{pendulum.now().format('MM/DD HH:mm:ss')}"
        event_desc = f"{event_descs[event_type]}"
        return " ".join([event_num, job_num, timestamp, event_desc])


class Job:
    """
    Class representing a faked HTCondor job
    """

    def __init__(self, file: Path, write_log: bool = True):
        self.fields, self.command = self.parse(file)
        self.write_log = write_log
        if self.write_log:
            self.logger = MockLogger(self.fields.get("log", "./condor.log"))
            self.logger.log_event(
                WorkflowEventType.SUBMITTED, f"Job node: {self.fields['executable']}"
            )

    def parse(self, file: Path) -> NamedTuple:
        """
        Parse condor job file for relevant fields
        :param file: Path to job file
        :return: NamedTuple of job data
        """
        # Parse file looking for executable and arguments
        JobData = namedtuple("JobData", ["field_data", "command"])
        section_sep = "\n\n"
        field_sep = "\n"

        with open(file, "r") as f:
            # Split file into data and action sections
            data, command = f.read().split(section_sep)
            fields = data.split(field_sep)
            field_dict = self.parse_fields(fields)
            return JobData(field_dict, command)

    @staticmethod
    def parse_fields(fields: List[str]) -> Dict[str, str]:
        """
        Take in list of fields parsed from input file and return them as a dictionary
        :param fields: List of fields
        :return: Dictionary of fields
        """
        field_dict = {}
        for field in fields:
            key, value = field.split(" = ")
            field_dict[key] = value
        return field_dict

    def execute(self):
        result_strings = {
            0: "Normal termination (return value 0)",
            # This is 255 instead of -1 because that's what subprocess returns instead of -1
            255: "Error in execution (return value -1)",
        }
        if self.write_log:
            self.logger.log_event(WorkflowEventType.EXECUTING)
        process = subprocess.run([self.fields["executable"], self.fields["arguments"]])
        exec_results = f"{result_strings.get(process.returncode, f'Something weird happened (return value {process.returncode})')}"

        if self.write_log:
            self.logger.log_event(WorkflowEventType.TERMINATED, exec_results)


class Dag:
    @staticmethod
    def parse(file):
        dag = Dag()
        # do something here to iterate the file contents and parse it
        raise NotImplementedError

    def execute(self):
        raise NotImplementedError
