from __future__ import annotations

import abc
import json
from enum import Enum, auto
from typing import Optional, List, Iterator

from .capability_interfaces import (
    ParameterIF,
    CapabilityStepIF,
    CapabilitySequenceIF,
    CapabilityEngineIF,
    CapabilityExecutionIF,
)
from .product_interfaces import FutureProductIF


class CapabilityStepType(Enum):
    """
    Enum that specifies the types of CapabilitySteps that are possible
    """

    PrepareAndRunWorkflow = 0
    AwaitQA = 1
    AwaitWorkflow = 2
    AwaitProduct = 3
    AwaitParameter = 4
    AwaitLargeAllocApproval = 5

    @classmethod
    def from_string(cls, string: str) -> CapabilityStepType:
        strings = {
            "prepare-and-run-workflow": cls.PrepareAndRunWorkflow,
            "await-qa": cls.AwaitQA,
            "await-workflow": cls.AwaitWorkflow,
            "await-product": cls.AwaitProduct,
            "await-parameter": cls.AwaitParameter,
            "await-large-alloc-approval": cls.AwaitLargeAllocApproval,
        }
        return strings[string]


class CapabilityStep(CapabilityStepIF, abc.ABC):
    """
    Class that represents a step in a capability sequence. A step is of a certain type and
    certain step types can have a value
    """

    TYPES = {}

    def __init__(
        self,
        step_type: CapabilityStepType,
        step_value: Optional[str],
        step_args: Optional[str],
    ):
        self.step_type = step_type
        self.step_value = step_value
        self.step_args = step_args

    @classmethod
    def from_str(cls, step_string: str):
        """
        Create CapabilityStep from string containing space-separated type, value, and args
        :param step_string: String of capability step, e.g. "PrepareAndRunWorkflow null"
        :return: CapabilityStep of given string
        """
        step_list = step_string.split(" ")
        step_type = CapabilityStepType.from_string(step_list[0])
        step_value = step_list[1] if step_list[1] else None
        step_args = step_list[2] if step_list[2] else None
        return cls.TYPES[step_type](step_type, step_value, step_args)

    def __str__(self):
        string = f"{self.step_type.name}"
        if self.step_value:
            string += f" {self.step_value}"
        elif self.step_args:
            string += f" {self.step_args}"
        return string


class CapabilitySequence(CapabilitySequenceIF):
    """
    Class that represents a capability sequence with a list of steps
    """

    def __init__(self, steps: List[CapabilityStep]):
        self.steps = steps

    @classmethod
    def from_str(cls, sequence_str: str):
        """
        Create CapabilitySequence from comma-separated str of capability steps
        :param sequence_str: String sequence of steps
        :return: CapabilitySequence of given steps
        """
        steps = []
        for step in sequence_str.split(","):
            steps.append(CapabilityStep.from_str(step))

        return cls(steps)

    def __iter__(self) -> Iterator:
        """
        Allows CapabilitySequence objects to be iterated over
        :return: Iterator object
        """
        return iter(self.steps)

    def __len__(self) -> int:
        """
        Allows len() to be called on CapabilitySequence objects
        :return: Number of capability steps
        """
        return len(self.steps)

    def __getitem__(self, step_number: int) -> CapabilityStep:
        return self.steps[step_number]

    def __str__(self) -> str:
        """
        Provides a string representation of the object
        :return: String representation
        """
        str_steps = [str(step) for step in self]
        return ",".join(str_steps)


class Parameter(ParameterIF):
    pass


class FutureProduct(FutureProductIF):
    pass


class PrepareAndRunWorkflow(CapabilityStep):
    def execute(self, engine: CapabilityEngineIF, execution: CapabilityExecutionIF):
        workflow_name = self.step_value
        workflow_args = self.step_args
        # FIXME: Add support for files
        files = []
        engine.submit_workflow_request(workflow_name, json.loads(workflow_args), files)


class AwaitQa(CapabilityStep):
    def execute(self, engine: CapabilityEngineIF, execution: CapabilityExecutionIF):
        pass


class AwaitWorkflow(CapabilityStep):
    def execute(self, engine: CapabilityEngineIF, execution: CapabilityExecutionIF):
        pass


class AwaitProduct(CapabilityStep):
    def execute(self, engine: CapabilityEngineIF, execution: CapabilityExecutionIF):
        pass


class AwaitParameter(CapabilityStep):
    def execute(self, engine: CapabilityEngineIF, execution: CapabilityExecutionIF):
        pass


class AwaitLargeAllocationApproval(CapabilityStep):
    def execute(self, engine: CapabilityEngineIF, execution: CapabilityExecutionIF):
        pass


CapabilityStep.TYPES = {
    CapabilityStepType.PrepareAndRunWorkflow: PrepareAndRunWorkflow,
    CapabilityStepType.AwaitQA: AwaitQa,
    CapabilityStepType.AwaitWorkflow: AwaitWorkflow,
    CapabilityStepType.AwaitProduct: AwaitProduct,
    CapabilityStepType.AwaitParameter: AwaitParameter,
    CapabilityStepType.AwaitLargeAllocApproval: AwaitLargeAllocationApproval,
}


class CapabilityEventType(Enum):
    """
    Enum that specifies the type of CapabilityEvents possible to send
    """

    WorkflowComplete = 0
    QaApproval = 1
    QaDenial = 1
    WorkflowReady = 2
    ProductReady = 3
    ParameterReady = 4
    LargeAllocApproval = 5
    LargeAllocDenial = 5


class WorkflowRequestState(Enum):
    """
    Enum that specifies the states a WorkflowRequest can find itself in
    """

    Complete = 0
    Created = 1
    Running = 2
    Failed = -1


class CapabilityRequestState(Enum):
    """
    Enum that specifies the states that a capability request can be in
    """

    Complete = 0
    Executing = 1
    Ready = 2
    Cancelled = 3


class ExecutionState(Enum):
    """
    Enum that specifies the states that a capability execution can find itself in
    """

    Complete = 0
    ExecutingStep = 1
    Ready = 2
    Cancelled = 3
    Failed = -1


class ExecutionPriority(Enum):
    """
    Enum that specifies the priority of a particular capability execution
    """

    High = 25
    Default = 50
    Low = 100
